
import random
from .__init__ import logger


class AbstractSearchOperator:

    def __init__(self, timeouts):
        """
        initializes the operator with a non-empty set of timeouts and an optional initial timeout
        """
        if len(timeouts) <= 0:
            raise ValueError('list of timeouts is empty')
        p = None
        for timeout in timeouts:
            if not 0 <= timeout:
                raise ValueError('timeout is negative')
            if p == None:
                p = timeout
            elif p >= timeout:
                raise ValueError('timeouts have to strictly increase')
        self.__timeouts = timeouts

        self.__current_index = 0
        self._timeout = self.__timeouts[self.__current_index]

    def increase_size(self):
        """
        increases the search time to the next defined timeout. 
        returns False if no increase is possible and True otherwise
        """
        if self.__current_index < len(self.__timeouts) - 1:
            old_timeout = self.__timeouts[self.__current_index]
            self.__current_index += 1
            self._timeout = self.__timeouts[self.__current_index]
        else:
            return False

        return True

    def decrease_size(self):
        """
        decreases the search time to the next defined timeout. 
        returns False if no decrease is possible and True otherwise
        """
        if self.__current_index > 0:
            self.__current_index -= 1
            self._timeout = self.__timeouts[self.__current_index]
        else:
            return False

        return True

    def reset_size(self):
        """
        resets the size of the operator to the lowest value
        """
        self.__current_index = 0
        self._timeout = self.__timeouts[self.__current_index]

    def flatten(self):
        """
        returns a list of operators where each contains only one of the timeouts
        """
        operators = []

        for timeout in self.__timeouts:
            operators += [type(self)(timeouts=[timeout])]

        return operators

    def execute(self, assumptions, time_left):
        pass

    def name(self):
        pass


class ClingoSearchOperator(AbstractSearchOperator):

    def __init__(self, internal_solver, name, timeouts, configuration, strict_bound_prob=1.0):
        super().__init__(timeouts)
        self.__name = name
        self.__timeouts = timeouts
        self.__configuration = configuration
        self.__strict_bound_prob = strict_bound_prob
        self.__internal_solver = internal_solver

    def execute(self, assumptions, time_left):
        timeout = min(self._timeout, time_left)
        logger.debug(
            f'operator "{self.__name}" executing search for {timeout} seconds')
        return self.__internal_solver.solve(timelimit=timeout, modellimit=1, assumptions=assumptions,
                                            configuration=self.__configuration, strict_bound=self.__strict_bound_prob > random.random())

    def flatten(self):
        """
        overriding to inject the internal solver
        """
        operators = []

        for timeout in self.__timeouts:
            operators += [ClingoSearchOperator(internal_solver=self.__internal_solver,
                                               name=self.__name, configuration=self.__configuration, timeouts=[timeout])]

        return operators

    def name(self):
        return f'{self.__name}: {self.__timeouts}'


# SearchOperator Factory

def get_operator(name, args, internal_solver):
    """
    returns a new search operator of the given type with given args
    """

    timeouts = args['timeouts']

    if 'configuration' in args:
        configuration = args['configuration']
    else:
        configuration = None
    if 'name' in args:
        name = args['name']
    else:
        name = None

    return ClingoSearchOperator(internal_solver, name, timeouts, configuration)



from __future__ import annotations

import hashlib
import importlib.util
import json
import logging
import os
import re
import subprocess
import sys
import time
import warnings
from collections import defaultdict
from contextlib import contextmanager
from copy import deepcopy
from datetime import datetime
from pathlib import Path
from subprocess import CompletedProcess
from typing import Callable, NoReturn

import dask_jobqueue
import dill
import numpy as np
from ConfigSpace import Configuration, ConfigurationSpace
from dask.distributed import Client
from dask_jobqueue import *
from sklearn.cluster import KMeans
from sklearn.exceptions import ConvergenceWarning
from sklearn.metrics import silhouette_score
from sklearn.preprocessing import StandardScaler
from smac import Scenario, HyperparameterOptimizationFacade
from smac.initial_design import DefaultInitialDesign
from smac.intensifier.intensifier import Intensifier
from smac.main.config_selector import ConfigSelector
from smac.runhistory import TrialInfo, TrialValue, StatusType, RunHistory

from .tuning_util import setupLogger, ensureFilesExist, readFile, updateDictAtPath, deleteCondition, resolveConfigurationSpace, resolveRelativePath, findAlaspoExecutable, parseInstance, SimpleNumpyJsonEncoder, \
    visualizeClustering, default

logger = logging.getLogger(__name__)
setupLogger(logger)


class TuningConfig:
    def __init__(
            self,
            timeLimit: int | None,
            trials: int | None,
            trialTimeLimit: int,
            encoding: str | list[str],
            instances: str,
            configurationSpace: dict | ConfigurationSpace,
            defaultConfig: dict,
            *,
            alaspoClArgs: list[str] = None,
            clusterInstances: bool | None = None,
            clusterSize: int | None = None,
            passInstanceFeaturesToSmac: bool | None = None,
            customFeatureExtractor: Callable[[str], list[float]] | None = None,
            standardizeInstanceFeatures: bool | None = None,
            visualizeClustering: bool | None = None,
            outputDir: str | None = None,
            workingDir: str | None = None,
            daskCluster: JobQueueCluster | None = None,
            daskWorkers: int | None = None
    ):
        requiredKeys = {
            "`timeLimit` or `trials`": timeLimit or trials,
            "`trialTimeLimit`": trialTimeLimit,
            "`encoding`": encoding,
            "`instances`": instances,
            "`configurationSpace`": configurationSpace,
            "`defaultConfig`": defaultConfig,
            "both `daskCluster` and `daskWorkers` or neither": bool(daskCluster) == bool(daskWorkers)
        }
        if not all(requiredKeys.values()):
            raise Exception(f"Missing one or more required config parameter(s): {', '.join(requiredKeys.keys())}")

        if daskCluster and timeLimit:
            raise Exception("Do not use timeLimit when using daskCluster! Specify the number of trials!")

        self.timeLimit: int | None = timeLimit
        self.trials: int | None = trials
        self.trialTimeLimit: int = trialTimeLimit
        self.workingDir: str = workingDir or os.path.abspath(os.getcwd())
        self.encoding: str | list[str] = self.resolvePath(encoding) if isinstance(encoding, str) else [self.resolvePath(x) for x in encoding]
        ensureFilesExist(self.encoding)
        instances = self.resolvePath(instances)
        ensureFilesExist(instances)
        with open(instances, "r") as f:
            actualInstances = [resolveRelativePath(os.path.dirname(instances), x.strip("\n \r")) for x in f.readlines()]
            ensureFilesExist(actualInstances)
        self.instances: list[str] = actualInstances
        self.nonPrimitives, self.configurationSpace = resolveConfigurationSpace(configurationSpace)
        self.config: dict = defaultConfig
        self.alaspoClArgs: list[str] = default(alaspoClArgs, sys.argv[1:])
        self.clusterInstances: bool = default(clusterInstances, True)
        self.clusterSize: int | None = clusterSize
        self.passInstanceFeaturesToSmac: bool = default(passInstanceFeaturesToSmac, True)
        self.customFeatureExtractor: Callable[[str], list[float]] | None = customFeatureExtractor
        self.standardizeInstanceFeatures: bool = default(standardizeInstanceFeatures, True)
        self.visualizeClustering: bool = default(visualizeClustering, False)
        self.outputDir: str | None = outputDir
        self.instanceFeatures: list[str] = []
        self.daskCluster: JobQueueCluster | None = daskCluster
        self.daskWorkers: int | None = daskWorkers

    def resolvePath(self, path: str) -> str:
        return resolveRelativePath(self.workingDir, path)

    @staticmethod
    def fromConfigPath(configPath: str) -> TuningConfig:
        config: dict = dict(json.loads(readFile(configPath)))
        tuningConf: dict = config["tuning"]
        configDir = os.path.dirname(configPath)
        return TuningConfig(
            timeLimit=tuningConf.get("timeLimit"),
            trials=tuningConf.get("trials"),
            trialTimeLimit=tuningConf.get("trialTimeLimit"),
            encoding=tuningConf.get("encoding"),
            instances=tuningConf.get("instances"),
            configurationSpace=tuningConf.get("configurationSpace"),
            defaultConfig=config,
            alaspoClArgs=tuningConf.get("alaspoClArgs"),
            clusterInstances=tuningConf.get("clusterInstances"),
            clusterSize=tuningConf.get("clusterSize"),
            passInstanceFeaturesToSmac=tuningConf.get("passInstanceFeaturesToSmac"),
            customFeatureExtractor=TuningConfig.getCustomFeatureExtractorFromFile(resolveRelativePath(configDir, tuningConf.get("customFeatureExtractor"))),
            standardizeInstanceFeatures=tuningConf.get("standardizeInstanceFeatures"),
            visualizeClustering=tuningConf.get("visualizeClustering"),
            outputDir=tuningConf.get("outputDir"),
            workingDir=tuningConf.get("workingDir", configDir),
            daskCluster=(d := TuningConfig.parseDaskConfig(tuningConf.get("dask")))[0],
            daskWorkers=d[1]
        )

    @staticmethod
    def parseDaskConfig(daskConfig: dict) -> tuple[JobQueueCluster, int] | tuple[None, None]:
        if not daskConfig or not (clusterType := getattr(dask_jobqueue, daskConfig.pop("type", ""), False)):
            return None, None
        daskWorkers: int = daskConfig.pop("workers")
        cluster: JobQueueCluster = clusterType(**daskConfig)
        return cluster, daskWorkers

    @contextmanager
    def excludeDaskCluster(self):
        daskCluster = self.daskCluster
        self.daskCluster = None
        try:
            yield
        finally:
            self.daskCluster = daskCluster

    @staticmethod
    def getCustomFeatureExtractorFromFile(filePath: str) -> Callable[[str], list[float]] | None:
        if not filePath:
            return None
        ensureFilesExist(filePath)
        filePath = os.path.abspath(filePath)
        spec = importlib.util.spec_from_file_location(os.path.splitext(os.path.basename(filePath))[0], filePath)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        try:
            return module.extractFeatures
        except AttributeError as e:
            raise Exception(f"No function named 'extractFeatures' found inside specified custom feature extractor at path '{filePath}'. "
                            "It has to have the signature 'extractFeatures(instancePath: str) -> list[float]'.") from e


###############################################################################################################################################################
###############################################################################################################################################################

class TunerFactory:
    @staticmethod
    def get(config: str | TuningConfig) -> ParameterTuner | DaskParameterTuner:
        c: TuningConfig = TuningConfig.fromConfigPath(config) if isinstance(config, str) else config
        if c.daskCluster:
            return DaskParameterTuner(c)
        return ParameterTuner(c)


###############################################################################################################################################################

class ParameterTuner:
    def __init__(self, config: str | TuningConfig):
        self.config: TuningConfig = TuningConfig.fromConfigPath(config) if isinstance(config, str) else config
        self.alaspoConfigs: dict[Configuration, dict[str, str]] = {}
        self.alaspoExecutable: str = findAlaspoExecutable()
        self.directory: str = self.config.outputDir or f"./tuning_{datetime.now().strftime('%Y-%m-%d_%H-%M-%S')}"
        os.makedirs(f"{self.directory}/alaspoConfigs/", exist_ok=True)
        self.results: dict[str, dict[str, float | None]] = defaultdict(lambda: defaultdict(lambda: None))

    def extractFeatures(self, instance: str) -> list[float]:  # does not handle e.g. clingo-dl - &diff directives (should be fine? since its instances; also, a custom extractor can be used)
        if self.config.customFeatureExtractor:
            res = self.config.customFeatureExtractor(instance)
            if not isinstance(res, list) or not all([isinstance(x, (float, int)) for x in res]):
                raise Exception(f"Custom feature extractor does not return list[float]. It returned {type(res)}: {res} instead.")
            return res
        counts: defaultdict = parseInstance(instance)
        if not self.config.instanceFeatures:  # take the features of first instance
            self.config.instanceFeatures = counts.keys()
        return [counts.get(f, 0) for f in self.config.instanceFeatures]

    @staticmethod
    def parseOutput(output: str) -> float:
        cost = None
        for line in output.splitlines():
            if "Cost" in line:
                c = line.split("Cost: ")[1]
                if c == "None":
                    continue
                cost = float(c)
        return cost if cost is not None else float('inf')

    def getAlaspoConfig(self, configuration: Configuration) -> str:
        if configuration in self.alaspoConfigs.keys():
            return self.alaspoConfigs[configuration]["path"]

        alaspoConfig = self.generateConfig(configuration)
        jsonDumps = json.dumps(alaspoConfig, indent=2, cls=SimpleNumpyJsonEncoder)
        path = os.path.join(self.directory, "alaspoConfigs", hashlib.sha256(jsonDumps.encode()).hexdigest() + ".json")

        with open(path, "w") as f:
            f.write(jsonDumps)
        self.alaspoConfigs[configuration] = {"path": path, "config": json.dumps(alaspoConfig, cls=SimpleNumpyJsonEncoder, separators=(",", ":"))}
        return path

    def generateConfig(self, configuration: Configuration) -> dict:
        alaspoConfig: dict = deepcopy(self.config.config)
        if "tuning" in alaspoConfig:
            del alaspoConfig["tuning"]
        for condition in configuration.config_space.conditions:
            deleteCondition(alaspoConfig, condition)  # cleanup conditional parameters (that may be contained in user-provided configuration)
        for k, v in configuration.items():
            v = self.config.nonPrimitives.get(v, v)  # resolve non-primitive types
            updateDictAtPath(k, v, alaspoConfig)
        return alaspoConfig

    def callAlaspo(self, configuration: Configuration, seed: int, instance: str) -> float:
        args = [x for arg in self.config.alaspoClArgs for x in (arg.split('=', 1) if '=' in arg else [arg])]
        if "--tuning-config" in args:
            idx = args.index("--tuning-config")
            args.pop(idx)
            args.pop(idx)

        alaspoConfig = self.getAlaspoConfig(configuration)
        cmd = f"{self.alaspoExecutable} {' '.join(args)} --seed {seed} --time-limit {self.config.trialTimeLimit} -c {alaspoConfig} -i {' '.join(self.config.encoding) if isinstance(self.config.encoding, list) else self.config.encoding} {instance}"
        logger.info(f"{configuration._values} - {instance}")
        process: CompletedProcess = subprocess.run(
            re.split("\s+|=", cmd),
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True
        )
        cost = self.parseOutput(process.stdout)
        logger.info(f"Cost: {cost}")
        self.results[instance][alaspoConfig] = cost
        return cost

    @staticmethod
    def kMeans(n_clusters: int, values: list[list[float]]) -> tuple[list[int], KMeans]:
        estimator: KMeans = KMeans(n_clusters=n_clusters, n_init="auto", random_state=42).fit(values)
        return estimator.predict(values), estimator

    def clusterInstances(self, featureMap: dict[str, list[float]]) -> dict[int, list[str]]:
        features = list(featureMap.values())
        instances = featureMap.keys()

        if not self.config.clusterInstances or len(instances) <= 2:
            with self.config.excludeDaskCluster():
                with open(os.path.join(self.directory, "clustering.dill"), "ab") as f:
                    dill.dump(None, f)
            return {0: self.config.instances}

        if self.config.clusterSize:
            labels, estimator = self.kMeans(self.config.clusterSize, features)
        else:  # use silhouette score to determine the best cluster size
            bestScore = -2
            warnings.filterwarnings(action="error", category=ConvergenceWarning)
            for i in range(2, len(instances)):
                try:
                    labels_, estimator_ = self.kMeans(i, features)
                except ConvergenceWarning:
                    break
                score = silhouette_score(features, labels_)
                if score > bestScore:
                    bestScore = score
                    labels = labels_
                    estimator = estimator_

        with self.config.excludeDaskCluster():
            with open(os.path.join(self.directory, "clustering.dill"), "ab") as f:
                dill.dump(estimator, f)

        clusters = defaultdict(list)
        for cluster, instance in zip(labels, instances):
            clusters[int(cluster)].append(instance)

        logger.info(f"Determined best number of clusters: {max(clusters.keys()) + 1}")
        return clusters

    @staticmethod
    def splitSimpleBudget(budget: int, clusters: int, clusterInstances: int, allInstances: int) -> float:
        return ((budget * 0.5) / clusters) + ((budget * 0.5 * clusterInstances) / allInstances)

    @staticmethod
    def splitTimeBudget(budget: int, clusters: dict[int, list[str]], allInstances: int) -> dict[int, float]:
        if not budget:
            return {c: np.inf for c in clusters.keys()}
        return {c: ParameterTuner.splitSimpleBudget(budget, len(clusters), len(i), allInstances) for c, i in clusters.items()}

    @staticmethod
    def splitTrialsBudget(budget: int, clusters: dict[int, list[str]], allInstances: int) -> dict[int, int]:
        if not budget:
            return {c: np.inf for c in clusters.keys()}
        trialsBudgetMap: dict[int, int] = {c: int(ParameterTuner.splitSimpleBudget(budget, len(clusters), len(i), allInstances)) for c, i in clusters.items()}
        diff: int = budget - sum(trialsBudgetMap.values())  # remaining trials lost due to truncation
        for cluster in sorted(clusters.keys(), key=lambda c: len(clusters[c]), reverse=True)[:diff]:
            trialsBudgetMap[cluster] += 1
        return trialsBudgetMap

    def tune(self) -> NoReturn:
        logger.info(f'Starting Hyperparameter Tuning')

        instanceFeatures: dict[str, list[float]] = {x: self.extractFeatures(x) for x in self.config.instances}
        scaler = None
        if self.config.standardizeInstanceFeatures:
            scaler = StandardScaler()
            standardizedFeatures = scaler.fit_transform(list(instanceFeatures.values())).tolist()
            instanceFeatures = dict(zip(instanceFeatures.keys(), standardizedFeatures))
        with self.config.excludeDaskCluster():
            with open(os.path.join(self.directory, "clustering.dill"), "wb") as f:
                dill.dump(scaler, f)
                dill.dump(self.extractFeatures, f)
        clusters: dict[int, list[str]] = self.clusterInstances(instanceFeatures)
        if self.config.visualizeClustering:
            visualizeClustering(
                clusterFeatures={k: [instanceFeatures[x] for x in v] for k, v in clusters.items()},
                clusterInstanceNames={k: [os.path.basename(x) for x in v] for k, v in clusters.items()},
                outputDir=self.directory
            )

        outFile: dict = {k: {"instances": v} for k, v in clusters.items()}

        timeBudgets: dict[int, float] = self.splitTimeBudget(self.config.timeLimit, clusters, len(self.config.instances))
        trialsBudgets: dict[int, int] = self.splitTrialsBudget(self.config.trials, clusters, len(self.config.instances))

        self.mainLoop(clusters, instanceFeatures, outFile, timeBudgets, trialsBudgets)

        with self.config.excludeDaskCluster():
            with open(os.path.join(self.directory, "clustering.dill"), "ab") as f:
                dill.dump(outFile, f)

        with open(os.path.join(self.directory, "results.json"), "w") as f:
            json.dump(outFile, f, indent=2, sort_keys=True)

        self.createCSV()

        logger.info(f"Finished tuning! Results written to {self.directory}")
        exit(0)

    def createCSV(self) -> None:
        with open(os.path.join(self.directory, "results.csv"), "w") as f:
            configPaths = [x["path"] for x in self.alaspoConfigs.values()]
            s = f"instance, {', '.join(configPaths)}\n"
            for instance, costMap in self.results.items():
                s += f"{instance},  {', '.join([str(costMap[config]) for config in configPaths])}\n"
            f.write(s)

    def mainLoop(self, clusters: dict[int, list[str]], instanceFeatures: dict[str, list[float]], outFile: dict, timeBudgets: dict[int, float], trialsBudgets: dict[int, int]) -> None:
        for cluster, instances in clusters.items():
            logger.info(f"Tuning cluster {cluster}")

            incumbent: Configuration = self.runSmac(timeBudgets[cluster], trialsBudgets[cluster], instances, {k: v for k, v in instanceFeatures.items() if k in instances})

            if not incumbent:  # too little time specified no experiment was run for this cluster
                raise Exception(f"The *timeLimit* specified is too little! Aborting!")

            bestConfig = self.alaspoConfigs[incumbent]
            outFile[cluster].update(bestConfig)
            logger.info(f"Best configuration for cluster {cluster}: {bestConfig['path']}")

    def runSmac(self, timeLimit: float, trialLimit: int, instances: list[str], instanceFeatures: dict[str, list[float]]) -> Configuration:
        scenario = Scenario(
            self.config.configurationSpace,
            deterministic=False,
            walltime_limit=timeLimit,
            n_trials=trialLimit,
            instances=instances,
            instance_features=instanceFeatures if self.config.passInstanceFeaturesToSmac else {instance: [i] for i, instance in enumerate(instances)},
            output_directory=Path(os.path.join(self.directory, "smac3_output"))
        )
        intensifier = Intensifier(scenario, max_config_calls=len(instances) * 5)  # set max_config_calls to make smac test on as many instances as it can (not just 3 by default) *with max 5 seeds per instance*
        initialDesign = DefaultInitialDesign(scenario)
        configSelector = ConfigSelector(scenario, retries=999)
        smac = HyperparameterOptimizationFacade(scenario, self.callAlaspo, intensifier=intensifier, initial_design=initialDesign, config_selector=configSelector)
        scenarioStartTime = time.time()
        trials = 0
        timeLimit: Callable[[], bool] = lambda: scenario.walltime_limit - (time.time() - scenarioStartTime) <= self.config.trialTimeLimit
        trialLimit: Callable[[], bool] = lambda: trials >= scenario.n_trials
        smacStop: Callable[[], bool] = lambda: (timeLimit() if self.config.timeLimit else False) or (trialLimit() if self.config.trials else False)
        while not smacStop():
            try:
                info: TrialInfo = smac.ask()
            except StopIteration:
                logger.info("No new configuration found to test.")
                break
            trialStartTime = time.time()
            cost = self.callAlaspo(info.config, info.seed, info.instance)
            trialEndTime = time.time()
            success: bool = cost is not None and not cost == float('inf')
            smac.tell(info, TrialValue(cost=cost, time=trialEndTime - trialStartTime, starttime=trialStartTime, endtime=trialEndTime, status=StatusType.SUCCESS if success else StatusType.CRASHED))
            trials += 1
        return smac.intensifier.get_incumbent()


###############################################################################################################################################################

class DaskParameterTuner(ParameterTuner):
    def mainLoop(self, clusters: dict[int, list[str]], instanceFeatures: dict[str, list[float]], outFile: dict, timeBudgets: dict[int, float], trialsBudgets: dict[int, int]) -> None:
        try:
            daskCluster: JobQueueCluster = self.config.daskCluster
            self.config.daskCluster = None  # remove unserializable object
            daskCluster.scale(self.config.daskWorkers)
            daskClient = Client(daskCluster)

            for cluster, instances in clusters.items():
                logger.info(f"Tuning cluster {cluster}")

                daskClient.wait_for_workers(1)  # wait until atleast one worker is ready
                incumbent: Configuration = self.runSmac(timeBudgets[cluster], trialsBudgets[cluster], instances, {k: v for k, v in instanceFeatures.items() if k in instances}, daskClient)

                if not incumbent:  # too little time specified no experiment was run for this cluster
                    raise Exception(f"The *timeLimit* specified is too little! Aborting!")

                bestConfig = {"path": (p := self.getAlaspoConfig(incumbent)), "config": readFile(p)}
                outFile[cluster].update(bestConfig)
                logger.info(f"Best configuration for cluster {cluster}: {bestConfig['path']}")
        finally:
            daskClient.shutdown()

    def runSmac(self, timeLimit: float, trialLimit: int, instances: list[str], instanceFeatures: dict[str, list[float]], daskClient: Client = None) -> Configuration:
        scenario = Scenario(
            self.config.configurationSpace,
            deterministic=False,
            walltime_limit=timeLimit,
            n_trials=trialLimit,
            instances=instances,
            instance_features=instanceFeatures if self.config.passInstanceFeaturesToSmac else {instance: [i] for i, instance in enumerate(instances)},
            output_directory=Path(os.path.join(self.directory, "smac3_output"))
        )
        intensifier = Intensifier(scenario, max_config_calls=len(instances) * 5)  # set max_config_calls to make smac test on as many instances as it can (not just 3 by default) *with max 5 seeds per instance*
        initialDesign = DefaultInitialDesign(scenario)
        configSelector = ConfigSelector(scenario, retries=999)
        smac = HyperparameterOptimizationFacade(scenario, self.callAlaspo, intensifier=intensifier, initial_design=initialDesign, dask_client=daskClient, config_selector=configSelector)
        incumbent: Configuration = smac.optimize()
        return incumbent

    def createCSV(self) -> None:
        if not os.path.exists(smacOutputDir := os.path.join(self.directory, "smac3_output")):
            return
        logger.info("Creating CSV file from SMAC3 runhistory ...")
        totalRh: RunHistory = RunHistory()
        for rhPath in Path(smacOutputDir).rglob("runhistory.json"):
            totalRh.update_from_json(rhPath, self.config.configurationSpace)
        self.alaspoConfigs = {c: {"path": self.getAlaspoConfig(c)} for c in totalRh.get_configs()}
        for k, v in totalRh.items():
            conf = totalRh.get_config(k.config_id)
            self.results[k.instance][self.alaspoConfigs[conf]["path"]] = v.cost
        super().createCSV()



import random
import sys
import os
import argparse
import signal
import logging
from . import lns
from . import solver
from . import initial
from . import strategy
from . import json_config
from . import relax
from . import search
from .__version__ import __version__
from .__init__ import logger

verbose = 0


def print_model(model):
    for a in model.shown:
        print(a, end=' ')
    print(" ")


def print_solution(solution, complete):
    if solution != None:
        if complete:
            print("Optimal solution:")
        else:
            print("Best found solution:")
        print_model(solution.model)
        print(f"Cost: {solution.cost}")
    else:
        print("No solution found")


def on_solution(solution, time):
    global verbose
    if verbose > 1:
        print('Solution:')
        print_model(solution.model)
        print(f'Time: {round(time, 3)}s')
    if verbose > 0:
        print(f'Cost: {solution.cost}')


def main():
    def existing_files(argument):
        if os.path.exists(argument) and os.path.isfile(argument):
            return argument
        else:
            raise argparse.ArgumentTypeError('File not found!')

    def valid_quick_config(argument):
        conf = argument.split(',')
        if len(conf) != 3:
            raise argparse.ArgumentTypeError('config string is not correct!')
        rate = float(conf[1])
        if not (0 <= rate <= 1):
            raise argparse.ArgumentTypeError('0 <= rate <= 1 required!')

        time = int(conf[2])
        if not (0 < time):
            raise argparse.ArgumentTypeError('timeout > 0 required!')

        return argument

    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                     description=f'ALASPO: ASP + Large-Neighborhood Search (version {__version__})')

    parser.add_argument('-i', '--input', type=existing_files, metavar='file', default=None, nargs='+',
                        help='input ASP files')

    parser.add_argument('-gt', '--time-limit', type=int, metavar='<n>', default=300,
                        help='time limit for the lns search')

    group = parser.add_mutually_exclusive_group()

    group.add_argument("-c", "--config-file", type=existing_files, metavar='<file>',
                       help='the config file specifying the relax and search operators')

    group.add_argument("-q", "--quick-config", type=valid_quick_config, metavar='<config>',
                       help='a config string containing a neighborhood type ("randomAtoms", "randomConstaints" or "declarative"), a relaxation rate, and a move timeout seperated by comma')

    tuning_group = parser.add_argument_group("tuning")
    tuning_group.add_argument("--tuning-config", type=existing_files, metavar='<config>',
                              help='the config file specifying the parameter tuning approach; for tuning documentation read examples/tuning/README.md')
    tuning_group.add_argument("--use-clustering", type=existing_files, metavar='<clustering.dill>',
                              help='the clustering.dill file specifying the the best configuration files per instance-cluster, resulting from a previous tuning run. Requires the use of "--instance"')
    tuning_group.add_argument('--instance', type=existing_files, metavar='<instance>',
                              help='the instance file which will determine the best configuration to be used. It will also be used as input ASP file')

    parser.add_argument('-st', '--solver-type', type=str, choices=['clingo', 'clingo-dl', 'clingcon'],
                        metavar='<arg>', default='clingo',
                        help='the ASP solver ("clingo", "clingo-dl", "clingcon") to be used')

    parser.add_argument('-mv', '--minimize-variable', type=str, metavar='<var>', default=None,
                        help='an integer variable to minimize (only useful with solver type "clingo-dl")')

    parser.add_argument('-sd', '--seed', type=int, metavar='SEED', default=None,
                        help='seed for random numbers')

    parser.add_argument('-ia', '--interactive', action='store_true',
                        help='select interactive selection strategy')
    parser.set_defaults(interactive=False)

    parser.add_argument('-v', '--verbosity', type=int, choices=[0, 1, 2],
                        metavar='<level>', default=1,
                        help='the level of verbosity in the output: 0 only output best found solution, 1 for updates about the incumbent cost, 2 for intermediate solutions')

    parser.add_argument('-d', '--debug', action='store_true',
                        help='enable debug logging')
    parser.set_defaults(debug=False)

    args = parser.parse_args()

    formatter = logging.Formatter(
        fmt='%(asctime)s - %(levelname)s - %(module)s - %(message)s')
    handler = logging.StreamHandler()
    handler.setFormatter(formatter)
    logger.addHandler(handler)

    global verbose
    verbose = args.verbosity

    if args.tuning_config:
        from . import tuning
        tuning.TunerFactory.get(args.tuning_config).tune()

    if args.use_clustering:
        if not args.instance:
            raise Exception('No instance file specified. Use "--instance" to specify the instance file.')
        if not args.input or not args.instance in args.input:
            args.input = (args.input or []) + [args.instance]

    if args.debug:
        logger.setLevel(level=logging.DEBUG)

    if args.seed is None:
        seed_value = random.randrange(sys.maxsize)
    else:
        seed_value = args.seed

    logger.info(f'Seed: {seed_value}')

    if verbose > 0:
        print(f'Seed: {seed_value}')

    random.seed(seed_value)

    program = ''
    if args.input is not None:
        for asp_file in args.input:
            with open(asp_file, 'r') as f:
                program += f.read()
    else:
        program += sys.stdin.read()

    internal_solver = None
    if args.solver_type == 'clingo':
        internal_solver = solver.Clingo(seed=seed_value)
    elif args.solver_type == 'clingo-dl':
        internal_solver = solver.ClingoDl(
            minimize_variable=args.minimize_variable, seed=seed_value)
    elif args.solver_type == 'clingcon':
        internal_solver = solver.Clingcon(seed=seed_value)
    else:
        raise ValueError("Not a valid solver type!")

    strat = None
    initial_operator = None
    relax_operators = None
    search_operators = None

    if args.config_file != None:
        with open(args.config_file, 'r') as f:
            con = f.read()
            strat, initial_operator, relax_operators, search_operators = json_config.parse_config(
                con, internal_solver)
    elif args.quick_config != None:
        conf_string = args.quick_config.split(',')
        op_name = conf_string[0].strip()
        rate = float(conf_string[1].strip())
        mt = int(conf_string[2].strip())

        strat = strategy.RandomStrategy(supports_intensification=True)
        relax_operators = [relax.get_operator(op_name, {'sizes': [rate]})]
        search_operators = [search.get_operator(
            'default', {'timeouts': [mt]}, internal_solver)]
    elif args.use_clustering:
        import dill
        with open(args.use_clustering, "rb") as f:
            scaler = dill.load(f)
            extractFeatures = dill.load(f)
            estimator = dill.load(f)
            configs = dill.load(f)
            if estimator:
                features = [extractFeatures(args.instance)]
                if scaler:
                    features = scaler.transform(features)
                cluster = estimator.predict(features)[0]
                bestConfig = configs[cluster]["config"]
                logger.info(f"Predicted instance to belong to cluster {cluster}")
            else:
                bestConfig = configs[0]["config"]
            logger.info(f"Determined the best configuration to be: {bestConfig}")
            strat, initial_operator, relax_operators, search_operators = json_config.parse_config(
                bestConfig, internal_solver)
    else:
        strat, initial_operator, relax_operators, search_operators = json_config.parse_config(
            json_config.DEFAULT_CONFIG, internal_solver)

    if initial_operator == None:
        initial_operator = initial.ClingoInitialOperator(internal_solver)

    # for interactive mode
    if args.interactive is True:
        strat = strategy.InteractiveStrategy()

    print('Solving...')
    lns_solver = lns.ClingoLNS(
        internal_solver, program, initial_operator, relax_operators, search_operators, strat)

    def signal_handler(sig, _):
        nonlocal lns_solver, strat
        sys.stderr.flush()
        sys.stdout.flush()
        print(f'Search interrupted! ({sig})')
        if verbose > 0:
            print_solution(lns_solver.best_solution,
                           lns_solver.search_complete)

        if sig != signal.SIGTERM and type(strat) == strategy.InteractiveStrategy:
            search_op = lns_solver.search_operator
            relax_op = lns_solver.relax_operator
            relax_operators, search_operators = lns_solver.get_portfolio()
            print(f'Current search configuration: {search_op.name()}')
            print('Search portfolio:')
            for i in range(len(search_operators)):
                print(f'{i}: {search_operators[i].name()}')
            while True:
                search = input(
                    'Select search configuration (ENTER = no change, C = exit): ')
                if search == '':
                    break
                if search.upper() == 'C':
                    sys.exit(0)
                if search.isdigit():
                    search_index = int(search)
                    if 0 <= search_index < len(search_operators):
                        search_op = search_operators[search_index]
                        break
                print('Not a valid index!')

            print(f'Current neighbourhood: {relax_op.name()}')
            print('Neighbourhood portfolio:')
            for i in range(len(relax_operators)):
                print(f'{i}: {relax_operators[i].name()}')
            while True:
                nh = input(
                    'Select search configuration (ENTER = no change, C = exit): ')
                if nh == '':
                    break
                if nh.upper() == 'C':
                    sys.exit(0)
                if nh.isdigit():
                    nh_index = int(nh)
                    if 0 <= nh_index < len(relax_operators):
                        relax_op = relax_operators[nh_index]
                        break
                print('Not a valid index!')

            strat.set_operators(relax_op, search_op)
        else:
            sys.exit(0)

    signal.signal(signal.SIGTERM, signal_handler)
    signal.signal(signal.SIGINT, signal_handler)

    solution = lns_solver.solve(args.time_limit, on_solution=on_solution)
    if verbose < 2:
        print_solution(solution, lns_solver.search_complete)
    elif lns_solver.search_complete:
        print('Found optimal solution.')


if __name__ == '__main__':
    main()



import argparse
import sys
import time
import clingo
import clingo.ast
import clingo.control
import clingo.theory
import clingodl
from clingo.symbol import Number
import clingcon
from .__init__ import logger


def apply_config_to_clingo(ctl, config):
    """ 
    helper function to apply a dict of settings directly to a clingo.configuration.
    """
    def apply_config(config, option, args):
        if type(args) == dict:
            for k, v in args.items():
                apply_config(getattr(config, option), k, v)
        elif type(args) == str:
            setattr(config, option, args)

    for k, v in config.items():
        apply_config(ctl.configuration, k, v)


class Clingo:
    """ 
    presents the the api of clingo.Control in a different way.
    """

    def __init__(self, *, options=None, seed=None, theory=None):

        clingoargs = []

        if seed:
            # clingo only accepts unsigned 32bit seed:
            cseed = seed & ((1 << 32) - 1)
            cseed = str(cseed)
            logger.debug("cseed: %s", cseed)
            clingoargs.append(f"--seed={cseed}")

        # the normal clingo.Control object is used. it is ground by the
        # initial operator and can then be used by the other operators.
        if options:
            clingoargs += options

        def my_logger(m, msg):
            if m != clingo.MessageCode.Other:
                print(msg, file=sys.stderr)

        self._ctl = clingo.control.Control(clingoargs, logger=my_logger)

        def extract_options(d, config):
            for k in config.keys:
                option = getattr(config, k)
                if type(option) == str:
                    d[k] = option
                elif type(option) == clingo.configuration.Configuration:
                    extract_options(d, option)

        self._init_config = {}
        extract_options(self._init_config, self._ctl.configuration)

        self._theory = theory
        if self._theory is not None:
            self._theory.register(self._ctl)

    def supports_native_opt(self):
        return True

    def _make_model(self, rawmodel):
        """helper method to make a long-lived copy of rawmodel.

        see clingo.Model, the return value somewhat mirrors it.
        """
        # logger.debug("rawmodel: %s", rawmodel) (this messes up output in interactive mode)
        model = argparse.Namespace()
        model.cost = rawmodel.cost.copy()
        model.number = rawmodel.number
        model.optimality_proven = rawmodel.optimality_proven
        model.thread_id = rawmodel.thread_id
        model.type = rawmodel.type
        model.symbols = list(rawmodel.symbols(
            atoms=True, terms=True, theory=True)).copy()
        model.shown = list(rawmodel.symbols(shown=True)).copy()
        model.assignments = {}

        return model

    def _read_cost(self, model):
        assert model
        if len(model.cost) == 0:
            return None
        elif len(model.cost) == 1:
            return model.cost[0]

        return model.cost

    def _make_solution(self, result, model):
        """makes a solution object

        Returns:
            a pickleable object, with attributes:
             - sat
                True ... the logic program is proven to be satisfiable
                False ... the logic program is proven to be un-satisfiable
                None ... satisfiability was not proven
             - cost
                None ... the cost is undefined / unknown
                int ... this solution's cost
             - model
                None ... no answer set was found
                Model ... the found answer set
        """
        sol = argparse.Namespace()
        sol.sat = result.satisfiable if result else None
        sol.cost = self._read_cost(model) if model else None
        sol.model = model
        sol.exhausted = result.exhausted if result else None

        return sol

    def _ast_visitor(self, ast, pb):
        """
        called on the addition of a new ast node to the program
        """
        pass

    def load_files(self, inputfiles):
        """
        see clingo.Control.load(str)
        """
        logger.debug("loading files: %s", inputfiles)

        with clingo.ast.ProgramBuilder(self._ctl) as pb:
            def callback(ast):
                self._ast_visitor(ast, pb)
                self._theory.rewrite_ast(
                    ast, pb.add) if self._theory else pb.add(ast)

            clingo.ast.parse_files(
                files=inputfiles,
                callback=callback)

    def load_string(self, inputstring):
        """
        see clingo.Control.load(str)
        """
        logger.debug("loading string")

        with clingo.ast.ProgramBuilder(self._ctl) as pb:
            def callback(ast):
                self._ast_visitor(ast, pb)
                self._theory.rewrite_ast(
                    ast, pb.add) if self._theory else pb.add(ast)

            clingo.ast.parse_string(
                inputstring,
                callback=callback)

    def ground(self):
        """
        see clingo.Control.ground([("base", [])])
        """
        logger.debug("grounding 'base'")
        self._ctl.ground([("base", [])])

    def _collect_models_on_model(self, rawmodel, models):
        if self._theory:
            self._theory.on_model(model=rawmodel)

        models.append(self._make_model(rawmodel))

    def _add_bound_less_than(self, bound):
        """
        adds the given bound(s) to the program
        """
        assert bound is not None
        assert type(bound) == list or type(bound) == int

        bound_eff = None
        if type(bound) == list:
            assert len(bound) > 0
            bound_eff = bound.copy()
            bound_eff[-1] -= 1
        else:
            bound_eff = [bound - 1]

        logger.debug('added bound: ' + 'opt, ' +
                     ', '.join([str(b) for b in bound_eff]))

        self._ctl.configuration.solve.opt_mode = 'opt, ' + \
            ', '.join([str(b) for b in bound_eff])

    def solve(self, assumptions=[], configuration=None, timelimit=None, modellimit=None, strict_bound=True):
        """
        Args:
          assumptions ... iterable of symbols to assume to be true. en
              empty array is equivalent to no assumptions.
          timelimit ... time limit in seconds. clingo.Control will be kindly
              asked to take no more than this amount of seconds to return.
              The time limit is not externally enforced.
              A value of ``None`` means no time limit.
          modellimit ... the maximal number of models returned by the solver
              (default = None)

        Returns:
            a solution or None if timelimit exceeded
        """
        # execute solve, and keep a copy of all models.
        models = []
        def on_model(rawmodel): return self._collect_models_on_model(
            rawmodel, models)
        assmpts = [(s, True) for s in assumptions]
        result = None

        # apply configuration
        if configuration != None:
            apply_config_to_clingo(self._ctl, configuration)
        else:
            apply_config_to_clingo(self._ctl, self._init_config)

        n_models = 0
        if modellimit is not None:
            n_models = modellimit

        self._ctl.configuration.solve.models = n_models

        with self._ctl.solve(
                async_=True,
                assumptions=assmpts,
                on_model=on_model
        ) as solveHandle:

            finished = solveHandle.wait(timelimit)
            if not finished:
                solveHandle.cancel()

            result = solveHandle.get()
            # print(result)
        if result is not None and not result.unknown:
            if not result.satisfiable:
                # solve() determined UNSAT
                assert 0 == len(models)
                return self._make_solution(result=result, model=None)
            else:
                # => some result&model was found within timelimit
                assert result.satisfiable is True
                assert 1 <= len(models)
                model = models[len(models) - 1]
                solution = self._make_solution(result=result, model=model)
                if solution.cost != None:
                    bound = None
                    if strict_bound:
                        bound = solution.cost
                    else:
                        bound = solution.cost + 1
                    self._add_bound_less_than(bound)

                return solution
        else:
            # timeout
            # Note: even if `models` contains some solution, it arrived
            # due to a race after the time was up.
            return self._make_solution(result=None, model=None)


class ClingoDl(Clingo):

    def __init__(self, *, options=None, seed=None, minimize_variable=None):

        self._theory = clingodl.ClingoDLTheory()

        super().__init__(options=options, seed=seed, theory=self._theory)

        self._minimize_variable = minimize_variable

        if self._minimize_variable:
            # if there is a minimization objective, we add the bound program module
            self._bound_id = 0
            part = f"#program bound(b). &diff {{ {self._minimize_variable} - 0 }} <= b."
            with clingo.ast.ProgramBuilder(self._ctl) as pb:
                clingo.ast.parse_string(
                    program=part,
                    callback=lambda ast: self._theory.rewrite_ast(ast, pb.add))

    def _make_model(self, rawmodel):
        model = super()._make_model(rawmodel)

        model.assignments = {}
        dl_assignments = []
        for dl_variable, dl_value in self._theory.assignment(rawmodel.thread_id):
            dl_assignments.append((str(dl_variable), dl_value))
        dl_assignments.sort()
        for dl_variable, dl_value in dl_assignments:
            model.assignments[str(dl_variable)] = dl_value

        return model

    def _read_cost(self, model):
        assert model
        cost = None
        if self._minimize_variable:
            for symb, value in model.assignments.items():
                if symb == self._minimize_variable:
                    cost = value
                    break
        else:
            logger.warning("no cost formula specified! (see options), "
                           "cost will be taken from clingo API.")
            cost = super()._read_cost(model)

        return cost

    def _add_bound_less_than(self, bound):
        """adds the given bound to the program

        the given bound is added to the logic program, such that the value of
        the minimize_variable (as named when invoking the constructor)
        MUST be LESS than the specified bound.

        this effect is currently  implemented via an additional difference
        constraint: ``&diff {variable - 0} <= boundeff.``
        In order to get "less-than" semantics while clingo-dl only supports
        the ``<=`` operator, the effective bound is: ``boundeff = bound - 1``
        """
        if self._minimize_variable:
            boundeff = bound - 1
            # ground new bound
            self._ctl.ground([('bound', [Number(boundeff)])])
            logger.debug('added bound: ' + str(boundeff))
        else:
            super()._add_bound_less_than(bound)

    def supports_native_opt(self):
        return False

    def solve(self, assumptions=[], configuration=None, timelimit=None, modellimit=None, strict_bound=True):
        start_time = time.time()
        def time_left(): return timelimit - (time.time() - start_time)

        if not self._minimize_variable:
            logger.debug('falling back to default clingo solve')
            # no clingo-dl minimization, defaulting to standard clingo solve
            return super().solve(assumptions=assumptions, timelimit=timelimit, modellimit=modellimit)
        else:

            if configuration != None:
                apply_config_to_clingo(self._ctl, configuration)
            else:
                apply_config_to_clingo(self._ctl, self._init_config)

            self._ctl.configuration.solve.models = 1

            models = []
            def on_model(rawmodel): return self._collect_models_on_model(
                rawmodel, models)
            assmpts = [(s, True) for s in assumptions]
            result = None
            solution = None

            result = None
            solution = None

            while (result is None or result.satisfiable) and (modellimit is None or len(models) < modellimit):

                with self._ctl.solve(
                        async_=True,
                        assumptions=assmpts,
                        on_model=on_model
                ) as solveHandle:

                    finished = solveHandle.wait(time_left())
                    if not finished:
                        solveHandle.cancel()

                    result = solveHandle.get()
                    # print(result)
                if result is not None and not result.unknown:
                    if not result.satisfiable:
                        # solve() determined UNSAT
                        assert 0 == len(models)
                        return self._make_solution(result=result, model=None)
                    else:
                        # => some result&model was found within timelimit
                        assert result.satisfiable is True
                        assert 1 <= len(models)
                        model = models[len(models) - 1]
                        solution = self._make_solution(
                            result=result, model=model)
                        if solution.cost != None:
                            bound = None
                            if strict_bound:
                                bound = solution.cost
                            else:
                                bound = solution.cost + 1
                            self._add_bound_less_than(bound)

                        return solution
                else:
                    # timeout
                    # Note: even if `models` contains some solution, it arrived
                    # due to a race after the time was up.
                    return self._make_solution(result=None, model=None)


class Clingcon(Clingo):

    def __init__(self, *, options=None, seed=None):

        self._theory = clingcon.ClingconTheory()

        super().__init__(options=options, seed=seed, theory=self._theory)

        self._minimize_atom = None

    def _make_model(self, rawmodel):
        model = super()._make_model(rawmodel)

        model.assignments = {}
        csp_assignments = []
        for csp_variable, csp_value in self._theory.assignment(rawmodel.thread_id):
            csp_assignments.append((str(csp_variable), csp_value))
        csp_assignments.sort()
        for csp_variable, csp_value in csp_assignments:
            model.assignments[str(csp_variable)] = csp_value

        return model

    def _read_cost(self, model):
        assert model

        for s in model.symbols:
            if s.match('__csp_cost', 1):
                return int(s.arguments[0].string)

        return super()._read_cost(model)

    def _ast_visitor(self, ast, pb):
        for item in ast.items():
            if item[0] == 'head':
                if item[1].ast_type == clingo.ast.ASTType.TheoryAtom:
                    if item[1].term.ast_type == clingo.ast.ASTType.Function:
                        if item[1].term.name == 'minimize':
                            if self._minimize_atom is not None:
                                raise Exception(
                                    'multiple minimize directives are currently not supported!')
                            self._minimize_atom = item[1]
                        elif item[1].term.name == 'maximize':
                            raise Exception(
                                'maximize directives are not supported yet!')

    def load_files(self, inputfiles):
        super().load_files(inputfiles)
        if self._minimize_atom:
            self.__add_bound_constraint()

    def load_string(self, inputstring):
        super().load_string(inputstring)
        if self._minimize_atom:
            self.__add_bound_constraint()

    def __add_bound_constraint(self):
        with clingo.ast.ProgramBuilder(self._ctl) as pb:
            pos = clingo.ast.Position('<string>', 1, 1)
            loc = clingo.ast.Location(pos, pos)
            program = clingo.ast.Program(
                loc, 'bound', [clingo.ast.Id(loc, 'b')])
            term = clingo.ast.Function(loc, 'sum', [], 0)
            sequence = list(self._minimize_atom.elements).copy()
            guard = clingo.ast.TheoryGuard('<=',
                                           clingo.ast.SymbolicTerm(loc, clingo.symbol.Function('b', [], True)))
            sum_atom = clingo.ast.TheoryAtom(loc, term, sequence, guard=guard)
            rule = clingo.ast.Rule(loc, sum_atom, [])
            self._theory.rewrite_ast(program, pb.add)
            self._theory.rewrite_ast(rule, pb.add)

    def _add_bound_less_than(self, bound):
        if self._minimize_atom:
            boundeff = bound - 1
            # ground new bound
            self._ctl.ground([('bound', [Number(boundeff)])])
            logger.debug('added bound %i' % boundeff)
        else:
            super()._add_bound_less_than(bound)




import random
from .__init__ import logger
import numpy as np
from itertools import chain, zip_longest

class AbstractStrategy():

    def prepare(self, relax_operators, search_operators):
        """
        prepares the strategy by providing the (non-empty) lists of relax and search operators.
        needs to be called before the strategy is used
        """
        if relax_operators is None or len(relax_operators) == 0:
            raise ValueError('there has to be at least one relax operator')

        if search_operators is None or len(search_operators) == 0:
            raise ValueError('there has to be at least one search operator')

        self._relax_operators = relax_operators
        self._search_operators = search_operators

    def get_portfolio(self):
        """
        returns a tuple containing the used relax and search operators
        """
        return (self._relax_operators, self._search_operators)

    def select_operators(self):
        """
        returns a pair of relax and search operator
        """
        pass

    def on_move_finished(self, operators, prev_cost, result, time_used):
        """
        called after the finish of a move to allow for statistics and adaptability
        """
        pass

    def supports_intensification(self):
        """
        whether or not the strategy supports intensification i.e. if the assumptions+operators are allowed to to be kept until no improvement can be achieved
        the default is no intensification
        """
        return False


class RandomStrategy(AbstractStrategy):

    def __init__(self, supports_intensification=False):
        self.__supports_intensification = supports_intensification

    def prepare(self, relax_operators, search_operators):
        super().prepare(relax_operators, search_operators)

        relax_operators = []
        for op in self._relax_operators:
            relax_operators += op.flatten()
        self._relax_operators = relax_operators

        search_operators = []
        for op in self._search_operators:
            search_operators += op.flatten()
        self._search_operators = search_operators

        logger.debug('random strategy selected')
        logger.debug('relax operators: ' +
                     str([o.name() for o in relax_operators]))
        logger.debug('search operators: ' +
                     str([o.name() for o in search_operators]))

    def select_operators(self):
        """
        returns a random pair of relax and search operator
        """
        relax_operator = random.choice(self._relax_operators)
        search_operator = random.choice(self._search_operators)

        return relax_operator, search_operator

    def supports_intensification(self):
        return self.__supports_intensification

class PowerLawStrategy(AbstractStrategy):

    def __init__(self, alpha=1.5, supports_intensification=False):
        self.__supports_intensification = supports_intensification
        self.__alpha = alpha

    def prepare(self, relax_operators, search_operators):
        super().prepare(relax_operators, search_operators)

        flattened_relax_operators = [ op.flatten() for op in relax_operators ]
        self._relax_operators = [ x for x in chain(*zip_longest(*flattened_relax_operators)) if x is not None ]

        flattened_search_operators = [ op.flatten() for op in search_operators ]
        self._search_operators = [ x for x in chain(*zip_longest(*flattened_search_operators)) if x is not None ]

        logger.debug('power law strategy selected')
        logger.debug('relax operators: ' +
                     str([o.name() for o in relax_operators]))
        logger.debug('search operators: ' +
                     str([o.name() for o in search_operators]))

    def __power_law_draw(self, values):
        indices = np.arange(1, len(values) + 1)
        probabilities = indices ** (-self.__alpha)
        probabilities /= probabilities.sum()  # Normalize to make it a valid probability distribution

        samples = np.random.choice(values, size=1, p=probabilities)

        return samples[0]

    def select_operators(self):
        """
        returns a pair of relax and search operator drawn using a power law distribution favouring lower indices
        """
        
        relax_operator = self.__power_law_draw(self._relax_operators)
        search_operator = self.__power_law_draw(self._search_operators)

        return relax_operator, search_operator

    def supports_intensification(self):
        return self.__supports_intensification


class DynamicStrategy(AbstractStrategy):

    def __init__(self, unsat_strike_limit=3, timeout_strike_limit=1):
        self.__unsat_strike_limit = unsat_strike_limit
        self.__timeout_strike_limit = timeout_strike_limit

    def prepare(self, relax_operators, search_operators):
        super().prepare(relax_operators, search_operators)

        self.__current_relax_operator = random.choice(self._relax_operators)
        self.__current_search_operator = random.choice(self._search_operators)

        self.__unsat_strikes = 0
        self.__timeout_strikes = 0

        logger.debug('variable strategy selected')
        logger.debug(
            f'relax operators: {[ o.name() for o in relax_operators ]}')
        logger.debug(
            f'search operators: {[ o.name() for o in search_operators ]}')

    def select_operators(self):
        return self.__current_relax_operator, self.__current_search_operator

    def on_move_finished(self, operators, prev_cost, result, time_used):
        if not result.sat:
            if result.exhausted:
                # UNSAT
                self.__unsat_strikes += 1
                if self.__unsat_strikes >= self.__unsat_strike_limit:
                    if not self.__current_relax_operator.increase_size():
                        self.__select_new_pair()
                    else:
                        self.__unsat_strikes = 0
                        logger.debug('increased relax size')

            else:
                # TIMEOUT
                self.__timeout_strikes += 1
                if self.__timeout_strikes >= self.__timeout_strike_limit:
                    if random.random() > 0.5:
                        # increase search time
                        if not self.__current_search_operator.increase_size():
                            self.__select_new_pair()
                        else:
                            self.__timeout_strikes = 0
                            logger.debug('increased search size')
                    else:
                        # reset relax size
                        self.__current_relax_operator.reset_size()
                        logger.debug('reset relax size')
        else:
            # IMPROVEMENT
            self.__unsat_strikes = 0
            self.__timeout_strikes = 0

    def supports_intensification(self):
        return True

    def __select_new_pair(self):
        logger.debug('selecting new operators')
        if len(self._relax_operators) > 1:
            relax_choices = [
                o for o in self._relax_operators if o != self.__current_relax_operator]
            self.__current_relax_operator = random.choice(relax_choices)

        self.__current_search_operator = random.choice(self._search_operators)

        self.__current_relax_operator.reset_size()
        self.__current_search_operator.reset_size()
        self.__unsat_strikes = 0
        self.__timeout_strikes = 0

        logger.debug(f'relax operator: {self.__current_relax_operator.name()}')
        logger.debug(
            f'search operator: {self.__current_search_operator.name()}')


class RouletteStrategy(AbstractStrategy):

    def __init__(self, alpha=0.5, lex_weight=1000):
        self.__alpha = alpha
        self.__lex_weight = lex_weight

    def prepare(self, relax_operators, search_operators):
        super().prepare(relax_operators, search_operators)

        relax_operators = []
        for op in self._relax_operators:
            relax_operators += op.flatten()
        self._relax_operators = relax_operators

        search_operators = []
        for op in self._search_operators:
            search_operators += op.flatten()
        self._search_operators = search_operators

        self._weights = {}

        for r_op in self._relax_operators:
            for s_op in self._search_operators:
                self._weights[(r_op, s_op)] = 1

        self._to_initialize = True

        logger.debug('roulette strategy selected')
        logger.debug(
            f'relax operators: {[ o.name() for o in relax_operators ]}')
        logger.debug(
            f'search operators: {[ o.name() for o in search_operators ]}')

    def select_operators(self):
        """
        returns a pair of relax and search operators depending on the weights
        """

        # logger.debug('weights: ' + str([ ((r.name(), s.name()), self._weights[(r,s)]) for r, s in self._weights ]))
        # logger.debug('cummulative sum of weights: ' + str(sum(self._weights.values())))

        weights = [float(w)/max(self._weights.values())
                   for w in self._weights.values()]
        relax_operator, search_operator = random.choices(
            list(self._weights.keys()), weights=weights, k=1)[0]

        logger.debug(f'selected relax operator: {relax_operator.name()}')
        logger.debug(f'selected search operator: {search_operator.name()}')

        return relax_operator, search_operator

    def on_move_finished(self, operators, prev_cost, result, time_used):
        cost = result.cost

        if cost is not None:
            if type(cost) == list:
                cost = self.calculate_weighted_sum(cost)

            if type(prev_cost) == list:
                prev_cost = self.calculate_weighted_sum(prev_cost)

            if self._to_initialize:
                for s_r_pair in self._weights:
                    self._weights[s_r_pair] = cost

                self._to_initialize = False

            ratio = (cost - prev_cost) / time_used
            self.update_weights(operators, ratio)

        else:
            self.update_weights(operators, 0)

        logger.debug(
            f'roulette weights: \n{[ ((r.name(), s.name()), self._weights[(r,s)]) for r, s in self._weights ]}')

    def update_weights(self, operators, ratio):
        new_weight = (1 - self.__alpha) * \
            self._weights[operators] - self.__alpha * ratio
        if new_weight < 0.001:
            new_weight = 0.001

        logger.debug(
            f'updating weight of {(operators[0].name(), operators[1].name())}: {self._weights[operators]} -> {new_weight}')
        self._weights[operators] = new_weight

    def calculate_weighted_sum(self, list):
        size = len(list) - 1
        cost = 0
        for i in range(len(list)):
            cost += list[i]*(self.__lex_weight**(size-i))
        return cost

# Strategy Factory


def get_strategy(type, args):
    """
    returns a new strategy of the given type
    """
    if type == 'random':
        return RandomStrategy()
    if type == 'powerlaw':
        alpha = 1.5
        if 'alpha' in args:
            alpha = args['alpha']
        return PowerLawStrategy(alpha=alpha)
    elif type == 'roulette':
        alpha = None
        if 'alpha' in args:
            alpha = args['alpha']
        lex_weight = None
        if 'lexWeight' in args:
            lex_weight = args['lexWeight']
        return RouletteStrategy(alpha=alpha, lex_weight=lex_weight)
    elif type == 'dynamic':
        unsat_strikes = None
        if 'unsatStrikes' in args:
            unsat_strikes = args['unsatStrikes']
        timeout_strikes = None
        if 'timeoutStrikes' in args:
            timeout_strikes = args['timeoutStrikes']
        return DynamicStrategy(unsat_strike_limit=unsat_strikes, timeout_strike_limit=timeout_strikes)
    else:
        raise ValueError(f"no strategy '{type}'")


class InteractiveStrategy(AbstractStrategy):

    def __init__(self, supports_intensification=False):
        self.__supports_intensification = supports_intensification

    def prepare(self, relax_operators, search_operators):
        super().prepare(relax_operators, search_operators)

        relax_operators = []
        for op in self._relax_operators:
            relax_operators += op.flatten()
        self._relax_operators = relax_operators

        search_operators = []
        for op in self._search_operators:
            search_operators += op.flatten()
        self._search_operators = search_operators

        self.__search_operator = search_operators[0]
        self.__relax_operator = relax_operators[0]

        logger.debug(
            'interactive strategy selected (Ctrl-C to interrupt search)')
        logger.debug(f'relax operators: {[o.name() for o in relax_operators]}')
        logger.debug(
            f'search operators: {[o.name() for o in search_operators]}')

    def select_operators(self):
        """
        returns current pair of relax and search operator
        """
        return self.__relax_operator, self.__search_operator

    def supports_intensification(self):
        return self.__supports_intensification

    def set_operators(self, relax_operator, search_operator):
        self.__relax_operator = relax_operator
        self.__search_operator = search_operator




import json
from . import strategy
from . import relax
from . import search
from . import initial

DEFAULT_CONFIG = """
{
    "strategy": {
        "name": "dynamic",
        "unsatStrikes": 3,
        "timeoutStrikes": 1
    },
    "relaxOperators": [
        {
            "type": "randomAtoms",
            "sizes": [ 0.1, 0.2, 0.4, 0.6, 0.8 ]
        },
        {
            "type": "randomConstants",
            "sizes": [ 0.1, 0.2, 0.3, 0.5 ]
        }
    ],
    "searchOperators": [
        {
            "name": "default",
            "timeouts": [ 5, 15, 30, 60 ]
        }
    ]
}
"""


def parse_config(config, internal_solver):
    json_config = json.loads(config)

    json_strategy = json_config['strategy']
    strat_name = json_strategy['name']
    strat_args = {k: v for k, v in json_strategy.items() if k != 'name'}

    strategy_op = strategy.get_strategy(strat_name, strat_args)

    relax_operators = []
    for json_relax in json_config['relaxOperators']:
        relax_type = json_relax['type']
        relax_args = {k: v for k, v in json_relax.items() if k != 'type'}
        relax_operators += [relax.get_operator(relax_type, relax_args)]

    search_operators = []
    for json_search in json_config['searchOperators']:
        search_name = json_search['name']
        search_args = {k: v for k, v in json_search.items() if k != 'type'}
        search_operators += [search.get_operator(
            search_name, search_args, internal_solver)]

    initial_operator = None
    if 'initialOperator' in json_config:
        json_initial = json_config['initialOperator']
        initial_args = {k: v for k, v in json_initial.items()}
        initial_operator = initial.get_operator(initial_args, internal_solver)

    return strategy_op, initial_operator, relax_operators, search_operators



from __future__ import annotations

import ast
import hashlib
import json
import logging
import os
import re
import shutil
import sys
from collections import defaultdict
from typing import TypeVar

import ConfigSpace
import numpy as np
from ConfigSpace import ConfigurationSpace
from clingo.ast import Transformer, AST, parse_files
from jsonpath_ng import JSONPath, parse

from .json_config import DEFAULT_CONFIG


def setupLogger(logger: logging.Logger) -> None:
    logger.setLevel(logging.DEBUG)
    stream_handler = logging.StreamHandler(sys.stdout)
    stream_handler.setLevel(logging.DEBUG)
    formatter = logging.Formatter(fmt='%(asctime)s - %(levelname)s - %(module)s - %(message)s')
    stream_handler.setFormatter(formatter)
    logger.addHandler(stream_handler)
    logger.propagate = False


def addDupPrefix(s: str, count: int) -> str:
    return f"#{count}_" + s


def removeDupPrefix(s: str) -> str:  # remove potential prefix that was added to avoid name conflicts (e.g., two strategies with equally named parameters)
    if not s.startswith("#"):
        return s
    return re.sub("^#\d+_", "", s)


class CfgPreproc:
    evalPrefix: str = "$eval:"

    @staticmethod
    def resolvePrefixes(item: dict | list) -> None:
        if isinstance(item, dict):
            CfgPreproc._resolveDict(item)
        elif isinstance(item, list):
            CfgPreproc._resolveList(item)

    @staticmethod
    def _resolveDict(dict_: dict) -> None:
        for key, val in dict_.items():
            dict_[key] = CfgPreproc._evalStringIfPrefixed(val)
            CfgPreproc.resolvePrefixes(dict_[key])

    @staticmethod
    def _resolveList(list_: list) -> None:
        for i, x in enumerate(list_):
            list_[i] = CfgPreproc._evalStringIfPrefixed(x)
            CfgPreproc.resolvePrefixes(list_[i])

    @staticmethod
    def _evalStringIfPrefixed(s: str | any) -> str | any:
        if isinstance(s, str) and s.startswith(CfgPreproc.evalPrefix):
            # return eval(s[len(self.evalPrefix):])
            return ast.literal_eval(s[len(CfgPreproc.evalPrefix):])
        return s

    @staticmethod
    def resolveNonPrimitives(dict_: dict) -> dict[str, object]:
        nonPrimitives: dict[str, object] = dict()
        for key, val in dict_.items():
            if not isinstance(val, list) or key == "strategy":
                continue
            for i, x in enumerate(val):
                if not isPrimitive(x):
                    hexdigest = sha256_hexdigest(str(x))
                    nonPrimitives[hexdigest] = x
                    val[i] = hexdigest
        return nonPrimitives


def isPrimitive(obj: any) -> bool:
    return isinstance(obj, (int, float, str, bool, type(None)))


def sha256_hexdigest(s: str) -> str:
    return hashlib.sha256(s.encode()).hexdigest()


def ConfigurationSpace_withDefaults(space: dict) -> ConfigurationSpace:  # set defaults for provided hyperparameters IF the default is valid
    defaultConfig = json.loads(DEFAULT_CONFIG)
    hyperparameters = list(ConfigurationSpace(space).values())

    for hp in hyperparameters:
        if d := parse(removeDupPrefix(hp.name)).find(defaultConfig):
            dVal = d[0].value
            if not isPrimitive(dVal):
                dVal = sha256_hexdigest(str(dVal))

            if hp.legal_value(dVal):  # i.e., check if the default value is either (1) within the provided choices for categorical hps or (2) inside the range of a uniform hp
                hp.default_value = dVal

    return ConfigurationSpace(hyperparameters)


def resolveConfigurationSpace(configConfigurationSpace: dict | ConfigurationSpace | None) -> tuple[dict[str, object], ConfigurationSpace]:
    if isinstance(configConfigurationSpace, ConfigurationSpace):
        return {}, configConfigurationSpace

    CfgPreproc.resolvePrefixes(configConfigurationSpace)
    nonPrimitives: dict[str, object] = CfgPreproc.resolveNonPrimitives(configConfigurationSpace)

    # resolve conditional parameters of strategies
    strategySelection: list[dict] = [s] if not isinstance(s := configConfigurationSpace.pop("strategy", []), list) else s
    if not strategySelection:
        return nonPrimitives, ConfigurationSpace_withDefaults(configConfigurationSpace)

    space = {"strategy.name": []}
    if configConfigurationSpace:
        space = {**space, **configConfigurationSpace}

    prefixCounts: dict[str, int] = defaultdict(lambda: 2)  # keep track of the number of occurrences of duplicate conditional parameters
    conditions: list[tuple[str, str, str]] = []
    for strat in strategySelection:
        stratName = strat["name"]
        space["strategy.name"].append(stratName)
        for k, v in strat.items():
            if k != "name":
                conditional = f"strategy.{k}"
                if conditional in space:  # equally named conditional hyperparameter exists
                    count = prefixCounts[conditional]
                    prefixCounts[conditional] += 1
                    conditional = addDupPrefix(conditional, count)
                space[conditional] = v
                conditions.append((conditional, "strategy.name", stratName))  # set ´conditional´ hyperparameter active if the value of the ´strategy.name´ hyperparameter equals ´stratName´

    cs = ConfigurationSpace_withDefaults(space)
    cs.add([ConfigSpace.conditions.EqualsCondition(cs[option], cs[strategy], strategyName) for option, strategy, strategyName in conditions])
    return nonPrimitives, cs


def updateDictAtPath(jsonPath: str, value, config: dict) -> None:
    path: JSONPath = parse(removeDupPrefix(jsonPath))
    path.update_or_create(config, value)


def deleteCondition(alaspoConfig: dict, condition: ConfigSpace.conditions.AbstractCondition) -> None:
    path: JSONPath = parse(removeDupPrefix(condition.child.name))
    path.filter(lambda _: True, alaspoConfig)


T = TypeVar('T')
U = TypeVar('U')


def default(obj: T | None, default: U) -> T | U:
    return obj if obj is not None else default


def readFile(filePath: str) -> str:
    with open(filePath, "r") as f:
        return f.read()


def ensureFilesExist(path_s: str | list[str]) -> None:
    if isinstance(path_s, str):
        path_s = [path_s]
    for path in path_s:
        if not os.path.exists(path) or os.path.isdir(path):
            raise FileNotFoundError(f"{path}")


def resolveRelativePath(directory: str, path: str) -> str | None:
    if path is None or os.path.isabs(path) or path.startswith("~"):
        return path
    return os.path.abspath(os.path.join(directory, path))


def findAlaspoExecutable() -> str:
    alaspoPath: str
    if os.name == 'nt':  # Windows
        alaspoPath = os.path.join(os.path.dirname(sys.executable), 'Scripts', 'alaspo.exe')
    else:
        alaspoPath = os.path.join(os.path.dirname(sys.executable), 'alaspo')

    if not os.path.exists(alaspoPath):  # not in env -> check PATH
        alaspoPath = shutil.which("alaspo")
        if not alaspoPath:
            return "alaspo"
    return alaspoPath


class AstVisitor(Transformer):
    def __init__(self):
        self.symbolicAtoms = defaultdict(lambda: 0)

    def visit_SymbolicAtom(self, ast: AST) -> AST:
        self.symbolicAtoms[ast.symbol.name] += 1
        return ast


def parseInstance(path: str) -> defaultdict:
    visitor = AstVisitor()
    parse_files([path], lambda stm: visitor(stm))
    return visitor.symbolicAtoms


class SimpleNumpyJsonEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, np.integer):
            return int(o)
        elif isinstance(o, np.floating):
            return float(o)
        elif isinstance(o, np.bool_):
            return bool(o)
        return super().default(o)


def visualizeClustering(clusterFeatures: dict[int, list[list[float]]], clusterInstanceNames: dict[int, list[str]], outputDir: str) -> None:
    from sklearn.decomposition import PCA
    import matplotlib.pyplot as plt
    import plotly.express as px

    labels, features, hoverInstanceNames = [], [], []

    for label, featuresList in clusterFeatures.items():
        features.extend(featuresList)
        labels.extend([label] * len(featuresList))
        hoverInstanceNames.extend(clusterInstanceNames[label])

    labels = np.array(labels)
    pcaFeatures_3d = PCA(n_components=3).fit_transform(features)
    pcaFeatures_2d = pcaFeatures_3d[:, :2]

    for label in clusterFeatures.keys():
        indices = labels == label
        plt.scatter(pcaFeatures_2d[indices, 0], pcaFeatures_2d[indices, 1], label=f'Cluster {label}')
    plt.title('Instance Clustering')
    plt.xlabel('PCA-1')
    plt.ylabel('PCA-2')
    plt.legend()
    plt.savefig(os.path.join(outputDir, 'clustering_plot_2d.pdf'))

    fig = px.scatter_3d(
        x=pcaFeatures_3d[:, 0],
        y=pcaFeatures_3d[:, 1],
        z=pcaFeatures_3d[:, 2],
        hover_name=hoverInstanceNames,
        color=labels.astype(str),
        title="Interactive Instance Clustering",
        labels={"x": "PCA-1", "y": "PCA-2", "z": "PCA-3", "color": "Cluster"}
    )
    fig.write_html(file=os.path.join(outputDir, "clustering_plot_3d.html"), include_plotlyjs=True)



from .__init__ import logger


class AbstractInititalOperator:

    def construct(self, global_timeout):
        """construct the initial solution used for lns

        Returns either:
            (1) a pickleable object, with attributes:
             - sat
                True ... the logic program is proven to be satisfiable
                False ... the logic program is proven to be un-satisfiable
                None ... satisfiability was not proven
             - cost
                None ... the cost is undefined / unknown
                int ... this solution's cost
             - model
                None ... no answer set was found
                Model ... the found answer set

            (2) a set of clingo.Symbols
        """
        pass


class ClingoInitialOperator(AbstractInititalOperator):

    def __init__(self, internal_solver, configuration=None, pre_opt_time=0):
        self.__internal_solver = internal_solver
        self.__configuration = configuration
        self.__pre_opt_time = pre_opt_time

    def construct(self, global_timeout):
        logger.debug('initial operator executing')
        if self.__pre_opt_time > 0:
            if type(self.__pre_opt_time) == float:
                t = round(global_timeout * self.__pre_opt_time)
            else:
                t = self.__pre_opt_time
            res = self.__internal_solver.solve(timelimit=min(t, global_timeout), configuration=self.__configuration)
            if not res.sat:
                time_left = global_timeout - min(t, global_timeout)
                res = self.__internal_solver.solve(timelimit=time_left, configuration=self.__configuration, modellimit=1)
            return res
        else:
            return self.__internal_solver.solve(timelimit=global_timeout, configuration=self.__configuration, modellimit=1)


# InitialOperator Factory

def get_operator(args, internal_solver):
    """
    returns a new initial operator of the given type with given args
    """

    if 'timeout' in args:
        timeout = args['timeout']
    else:
        timeout = 0

    if 'configuration' in args:
        configuration = args['configuration']
    else:
        configuration = None

    return ClingoInitialOperator(internal_solver, configuration, timeout)



import time
from . import initial
import clingo
from .__init__ import logger


class ClingoLNS:

    def __init__(self, internal_solver, program, initial_operator, relax_operators, search_operators, strategy):
        """
        instantiated the VLNS solver with required relax operators and the optional move timeout in seconds (default 5)
        """
        self.__internal_solver = internal_solver
        self.__program = program
        self._unsat_count = 0
        self._timeout_count = 0

        if initial_operator is None:
            raise ValueError('no initial operator provided')

        self.__initial_operator = initial_operator

        self.__strategy = strategy

        # keep references current operators
        self.relax_operator = None
        self.search_operator = None

        self.__strategy.prepare(relax_operators, search_operators)

        self.best_solution = None
        self.search_complete = False

    def get_portfolio(self):
        """
        returns a tuple containing the used relax and search operators
        """

        return self.__strategy.get_portfolio()

    def solve(self, timeout, on_solution=None):
        """
        runs the VLNS algorithm on the given ASP instance for the given timelimit
        """

        self._unsat_count = 0
        self._timeout_count = 0

        start_time = time.time()
        def time_left(): return timeout - (time.time() - start_time)

        # get internal solver
        internal_solver = self.__internal_solver

        # load clear and program
        internal_solver.load_string(self.__program)

        # ground base
        internal_solver.ground()

        incumbent = None

        # obtain initial solution
        solution = self.__initial_operator.construct(time_left())

        if solution is None or solution == []:
            logger.info('NO INTITIAL SOLUTION FOUND')
            print('Could not obtain initial solution!')
            return None

        if isinstance(solution, list) and all(isinstance(x, clingo.symbol.Symbol) for x in solution):
            # non default init operator provided a set of symbols which need to be feed to the solver
            solution = internal_solver.solve(
                assumptions=solution, modellimit=1, timelimit=time_left())
        elif not isinstance(self.__initial_operator, initial.ClingoInitialOperator):
            # non default init operator was used, hence we seed the solver with the greedy solution
            internal_solver.solve(
                assumptions=solution.model.symbols, modellimit=1, timelimit=time_left())

        logger.info(f'initial cost: {solution.cost}')
        incumbent = solution
        if on_solution != None and solution != None:
            on_solution(incumbent, time.time() - start_time)
        self.best_solution = incumbent

        if solution.exhausted:
            logger.info('OPTIMAL SOLUTION FOUND')
            self.search_complete = True
            return incumbent

        # LNS loop
        assumptions = None
        while time_left() > 0:
            move_start_time = time.time()

            # get assumptions
            if assumptions is None or not self.__strategy.supports_intensification():
                self.relax_operator, self.search_operator = self.__strategy.select_operators()
                logger.debug('selected relax operator %s and search operator %s' % (
                    self.relax_operator.name(), self.search_operator.name()))
                assumptions = self.relax_operator.get_move_assumptions(
                    incumbent)
            # perform move
            max_move_time = time_left()
            if max_move_time <= 0:
                break
            solution = self.search_operator.execute(assumptions, max_move_time)

            prev_cost = incumbent.cost
            if solution.sat:
                # solution found, update incumbent
                incumbent = solution
                logger.info(f'found solution with cost: {incumbent.cost}')
                if solution.exhausted:
                    logger.debug('move optimal')
                if on_solution != None:
                    on_solution(incumbent, time.time() - start_time)
                self.best_solution = incumbent
                if prev_cost == solution.cost:
                    assumptions = None
                self._unsat_count = 0
                self._timeout_count = 0
            else:
                # unsat or timeout, do not change incumbent and reset assumptions
                if solution.sat is False or solution.exhausted:
                    self._timeout_count = 0
                    if len(assumptions) == 0:
                        logger.info('OPTIMAL SOLUTION FOUND')
                        self.search_complete = True
                        return incumbent
                    else:
                        logger.debug('unsat/optimal under current assumptions')
                        self._unsat_count += 1
                else:
                    logger.debug('move timed out')
                    self._unsat_count = 0
                    self._timeout_count += 1
                assumptions = None

            move_end_time = time.time()
            operators = (self.relax_operator, self.search_operator)
            self.__strategy.on_move_finished(
                operators, prev_cost, solution, move_end_time - move_start_time)

        return incumbent




import random
from . import config
from .__init__ import logger


class AbstractRelaxOperator():

    def __init__(self, sizes=[1]):
        """
        initializes the operator with a non-empty set of relaxation sizes and an optional initial size. 
        the sizes are either all relative (between zero and one) or all absolute (integers bigger than zero)
        """
        if len(sizes) <= 0:
            raise ValueError('list of sizes is empty')

        absolute = None
        self._sizes = sorted(sizes)
        relative = True
        for size in sizes:
            if not (size > 0.0 and size < 1.0):
                relative = False
                break

        absolute = True
        for size in sizes:
            if not (size > 0 and (type(size) == int or (type(size) == float and size.is_integer()))):
                absolute = False
                break

        if not (absolute or relative):
            raise ValueError('sizes have to be all relative or all absolute!')

        self.__current_index = 0

        self._size = self._sizes[self.__current_index]
        self._absolute = absolute

    def get_move_assumptions(self, incumbent):
        """
        returns the assumptions for the next VLNS move i.e. which parts of the
        solution are not relaxed according to the given rate
        """
        pass

    def increase_size(self):
        """
        increases the relaxation rate to the next defined rate. 
        returns False if no increase is possible and True otherwise
        """
        if self.__current_index < len(self._sizes) - 1:
            self.__current_index += 1
            self._size = self._sizes[self.__current_index]
        else:
            return False

        return True

    def decrease_size(self):
        """
        decreases the relaxation rate to the next defined rate. 
        returns False if no decrease is possible and True otherwise
        """
        if self.__current_index > 0:
            self.__current_index -= 1
            self._size = self._sizes[self.__current_index]
        else:
            return False

        return True

    def reset_size(self):
        """
        resets the size of the operator to the lowest value
        """
        self.__current_index = 0
        self._size = self._sizes[self.__current_index]

    def minimize_size(self):
        """
        decreases the relaxation rate to the next defined rate. 
        returns False if no decrease is possible and True otherwise
        """
        if self.__current_index > 0:
            self._size = self._sizes[0]
        else:
            return False

        return True

    def name(self):
        """
        return a string identifier for the operator (used for logging and statistics)
        """
        pass

    def flatten(self):
        """
        returns a list of operators where each contains only one of the rates
        """
        operators = []

        if len(self._sizes) > 1:
            for size in self._sizes:
                operators += [type(self)(sizes=[size])]
        else:
            operators = [self]

        return operators


class RandomAtomRelaxOperator(AbstractRelaxOperator):

    def __init__(self, sizes):
        super().__init__(sizes)

    def get_move_assumptions(self, incumbent):
        asm = []

        max_selection_sz = len(incumbent.model.shown)

        if self._absolute:
            selection_sz = min(max_selection_sz, self._size)
            selection_sz = max_selection_sz - selection_sz
        else:
            selection_sz = round(max_selection_sz * (1 - self._size))

        asm = random.sample(incumbent.model.shown, selection_sz)

        logger.debug(
            f'atom operator relaxed {max_selection_sz - selection_sz} / {max_selection_sz} atoms.')

        return asm

    def name(self):
        return 'random atoms: ' + str(self._sizes)


class RandomConstantRelaxOperator(AbstractRelaxOperator):

    def __init__(self, sizes):
        super().__init__(sizes)

    def get_move_assumptions(self, incumbent):
        constants = set()
        for s in incumbent.model.shown:
            constants = constants.union(s.arguments)

        if self._absolute:
            relaxed_number = min(len(constants), self._size)
        else:
            relaxed_number = int(len(constants) * self._size)
        relaxed_constants = set(random.sample(
            sorted(constants), relaxed_number))

        assumptions = [
            s for s in incumbent.model.shown if relaxed_constants.isdisjoint(s.arguments)]

        logger.debug(f"constant operator relaxed "
                     f"{len(incumbent.model.shown) - len(assumptions)} / {len(incumbent.model.shown)} atoms.")

        return assumptions

    def name(self):
        return 'random constants: ' + str(self._sizes)


class DeclarativeRelaxOperator(AbstractRelaxOperator):

    def __init__(self, sizes, name=None):
        super().__init__(sizes)
        self.__name = name

    def get_move_assumptions(self, incumbent):
        """
        returns the assumptions for the next VLNS move i.e. which parts of the
        solution are not relaxed according to the given rate
        """
        asm = []

        select = []
        fix = []

        for s in incumbent.model.symbols:

            if self.__name is None:
                if s.match(config.SELECT_PRED, 1):
                    select.append(s.arguments[0])

                elif s.match(config.FIX_PRED, 2):
                    fix.append((s.arguments[0], s.arguments[1]))

            else:
                if s.match(config.SELECT_PRED, 2) and s.arguments[0].name == self.__name:
                    select.append(s.arguments[1])

                elif s.match(config.FIX_PRED, 3) and s.arguments[0].name == self.__name:
                    fix.append((s.arguments[1], s.arguments[2]))

        max_selection_sz = len(select)
        if max_selection_sz <= 0:
            raise ValueError('empty selection')

        if self._absolute:
            selection_sz = min(max_selection_sz, self._size)
            selection_sz = max_selection_sz - selection_sz
        else:
            selection_sz = round(max_selection_sz * (1 - self._size))

        selection = random.sample(select, selection_sz)
        for sel in selection:
            for atom, s in fix:
                if sel == s:
                    asm.append(atom)

        logger.debug(f"lns_select operator relaxed "
                     f"{max_selection_sz - selection_sz} / {max_selection_sz} atoms.")

        return asm

    def name(self):
        if self.__name != None:
            return 'lns_select "' + self.__name + '": ' + str(self._sizes)
        else:
            return 'lns_select: ' + str(self._sizes)

    def flatten(self):
        """
        returns a list of operators where each contains only one of the rates
        """
        operators = []

        for size in self._sizes:
            operators += [type(self)(sizes=[size], name=self.__name)]

        return operators

# RelaxOperator Factory


def get_operator(type, args):
    """
    returns a new relax operator of the given type with given args
    """
    sizes = args['sizes']

    if type == 'randomAtoms':
        return RandomAtomRelaxOperator(sizes)
    elif type == 'randomConstants':
        return RandomConstantRelaxOperator(sizes)
    elif type == 'declarative':
        name = None
        if 'name' in args:
            name = args['name']

        return DeclarativeRelaxOperator(sizes, name=name)
    else:
        raise ValueError('unknown relax operator "%s"' % type)
